# Copyright 2021 Okera Inc. All Rights Reserved.
#
# Some scenario tests for steward delegation
#
# pylint: disable=broad-except
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=bad-continuation
# pylint: disable=broad-except

"""
+---------------+--------------------------+----------+---------+---------+---------+---------+-------+
|      rpc      |           name           |   user   |   mean  |   50%   |   90%   |   99%   | iters |
+---------------+--------------------------+----------+---------+---------+---------+---------+-------+
| list_datasets | Wide table w/ attributes |  admin   |  57.408 |  56.484 |  57.609 |  93.027 |   50  |
| list_datasets | Wide table w/ attributes | testuser | 334.477 | 312.530 | 373.459 | 577.000 |   50  |
+---------------+--------------------------+----------+---------+---------+---------+---------+-------+
"""

import statistics

from okera import context
from okera import _thrift_api
from okera.tests import pycerebro_test_common as common
from prettytable import PrettyTable

ATTR = "perf_test.attr1"
DB = "python_perf_test_db"
TBL1 = "tbl1"
TBL2 = "tbl2"
ROLE = "python_perf_test_role"
TEST_USER = "pythonperfuser"

def _create_wide_view(db, tbl, multiple, source_db, source_tbl):
    cols = []
    for idx in range(multiple):
        cols.extend([
            "uid AS uid%04d" % (idx),
            "dob AS dob%04d" % (idx),
            "gender AS gender%04d" % (idx),
            "ccn AS ccn%04d" % (idx),
        ])
    stmt = "CREATE VIEW %s.%s AS SELECT %s FROM %s.%s" % (
        db, tbl, ', '.join(cols), source_db, source_tbl)

    return stmt

OUTPUT_TABLE = PrettyTable(
  ['rpc', 'name', 'user', 'mean', '50%', '90%', '99%', 'iters']
)

class PerfTest(common.TestBase):
    @classmethod
    def setUpClass(cls):
        print("\n\nSetting up perf test.")
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ddls = [
                "DROP ATTRIBUTE IF EXISTS %s" % (ATTR),
                "CREATE ATTRIBUTE %s" % (ATTR),

                "DROP DATABASE IF EXISTS %s CASCADE" % (DB),
                "CREATE DATABASE %s" % (DB),

                """CREATE TABLE {db}.users (
                  uid STRING ATTRIBUTE {attr},
                  dob STRING ATTRIBUTE {attr},
                  gender STRING ATTRIBUTE {attr},
                  ccn STRING ATTRIBUTE {attr}
                )""".format(db=DB, attr=ATTR),

                _create_wide_view(DB, TBL1, 100, DB, 'users'),

                "DROP ROLE IF EXISTS %s" % (ROLE),
                "CREATE ROLE %s WITH GROUPS %s" % (ROLE, TEST_USER),
                "GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE IN (%s) TO ROLE %s" % \
                    (DB, ATTR, ROLE),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)
        print("...Done setting up perf test")

    @classmethod
    def tearDownClass(cls):
        print("\nPerf tests done:")
        print(str(OUTPUT_TABLE))

    def measure(self, fn, msg, user, rpc, iters=50):
      latencies = common.measure_latency(iters, fn, msg=msg)
      mean = "{:.3f}".format(statistics.mean(latencies))
      p50 = "{:.3f}".format(latencies[int(len(latencies) * .5)])
      p90 = "{:.3f}".format(latencies[int(len(latencies) * .9)])
      p99 = "{:.3f}".format(latencies[int(len(latencies) * .99)])
      row = [rpc, msg, user, mean, p50, p90, p99, iters]
      OUTPUT_TABLE.add_row(row)

    def test_wide_table_with_attributes(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            def list(tbl):
                def get():
                    datasets = conn.list_datasets(DB, name=tbl)
                    assert len(datasets) == 1
                return get

            """
            ---------------------------- admin --------------------------
            Iterations 50
            Mean 57.408270835876465 ms
            50%: 56.48446083068848 ms
            90%: 57.608842849731445 ms
            95%: 58.23540687561035 ms
            99%: 93.02711486816406 ms
            99.5%: 93.02711486816406 ms
            99.9%: 93.02711486816406 ms
            """
            self.measure(list(TBL1), "Wide table w/ attributes",
                         "admin", "list_datasets")

            """
            ---------------------------- test user --------------------------
            Iterations 50
            Mean 334.47683811187744 ms
            50%: 312.5302791595459 ms
            90%: 373.4593391418457 ms
            95%: 374.4511604309082 ms
            99%: 576.9996643066406 ms
            99.5%: 576.9996643066406 ms
            99.9%: 576.9996643066406 ms
            """
            ctx.enable_token_auth(token_str=TEST_USER)
            self.measure(list(TBL1), "Wide table w/ attributes",
                         "testuser", "list_datasets")




# Copyright 2017 Okera Inc. All Rights Reserved.
#
# Tests some queries against nightly
import os
import unittest

from okera import context
from okera.tests import pycerebro_test_common as common

ROOT_TOKEN = os.environ['OKERA_HOME'] + '/integration/tokens/cerebro.token'
NIGHTLY_PLANNER = 'nightly.internal.okera.rocks'

class ConnectionErrorsTest(unittest.TestCase):

    @unittest.skipIf(common.TEST_LEVEL == 'smoke', "Skipping at unit test level.")
    def test_okera_sample_users(self):
        ctx = context()
        ctx.enable_token_auth(token_file=ROOT_TOKEN)
        with ctx.connect(NIGHTLY_PLANNER) as conn:
            json_data = conn.scan_as_json('okera_sample.users')
            self.assertEqual(38455, len(json_data))
            json_data = conn.scan_as_json('okera_sample.users', max_records=200)
            self.assertEqual(200, len(json_data))
            json_data = conn.scan_as_json('okera_sample.users', max_records=40000)
            self.assertEqual(38455, len(json_data))

            json_data = conn.scan_as_json('okera_sample.users',
                                          max_records=200,
                                          max_client_process_count=2)
            self.assertEqual(200, len(json_data))
            pd = conn.scan_as_pandas('okera_sample.users')
            self.assertEqual(38455, len(pd))
            pd = conn.scan_as_pandas('okera_sample.users', max_records=200)
            self.assertEqual(200, len(pd))
            pd = conn.scan_as_pandas('okera_sample.users', max_records=40000)
            self.assertEqual(38455, len(pd))
            pd = conn.scan_as_pandas('okera_sample.users',
                                     max_records=200,
                                     max_client_process_count=2)
            self.assertEqual(200, len(pd))

    @unittest.skipIf(common.TEST_LEVEL == 'smoke', "Skipping at unit test level.")
    def test_okera_fs_read(self):
        common.configure_botocore_patch()
        import boto3
        s3 = boto3.client('s3')
        res = s3.get_object(Bucket='cerebrodata-test', Key='large-doubles/data.txt')
        output = res['Body'].read().decode('UTF-8').strip(' \t\n\r')
        self.assertEqual(output, '901225911454.29')

if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Okera Inc. All Rights Reserved.
#
# Some scenario tests for steward delegation
#
# pylint: disable=broad-except
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=bad-continuation
# pylint: disable=broad-except


import os
import statistics
import time

from okera import context
from okera import _thrift_api
from okera.tests import pycerebro_test_common as common
from prettytable import PrettyTable

def read_sql_file(path, **kwargs):
    sql = open(path, 'r').read()
    return sql.format(**kwargs)

class XilinxTest(common.TestBase):

    def test_xilinx_tbl1(self):
        db = 'xilinx_test_db'
        tbl1 = 'flash_view_wtd'
        role = 'xilinx_test_role'
        testuser = 'xilinxuser'

        flash_view_wtd_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            'assets', 'xilinx_flash_view_wtd.sql')

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ddls = [
                "DROP DATABASE IF EXISTS %s CASCADE" % (db),
                "CREATE DATABASE %s" % (db),
                read_sql_file(flash_view_wtd_path, db=db, tbl=tbl1),

                "create attribute if not exists flash.cogs_bklg_cra",
                "create attribute if not exists flash.cogs_bklg_si",
                "create attribute if not exists flash.cogs_bklg_st",
                "create attribute if not exists flash.cogs_hist_cra",
                "create attribute if not exists flash.cogs_hist_si",
                "create attribute if not exists flash.cogs_hist_st",
                "create attribute if not exists flash.dimension",
                "create attribute if not exists flash.measure",
                "create attribute if not exists flash.qty_bklg_cra",
                "create attribute if not exists flash.qty_bklg_si",
                "create attribute if not exists flash.qty_bklg_st",
                "create attribute if not exists flash.qty_hist_cra",
                "create attribute if not exists flash.qty_hist_si",
                "create attribute if not exists flash.qty_hist_st",
                "create attribute if not exists flash.resale_bklg_st",
                "create attribute if not exists flash.resale_hist_st",
                "create attribute if not exists flash.revenue_bklg_cra",
                "create attribute if not exists flash.revenue_bklg_si",
                "create attribute if not exists flash.revenue_bklg_st",
                "create attribute if not exists flash.revenue_hist_cra",
                "create attribute if not exists flash.revenue_hist_si",
                "create attribute if not exists flash.revenue_hist_st",
                "create attribute if not exists flash.revenue_cover_st",
                "create attribute if not exists flash.asp_hist_st",
                "create attribute if not exists flash.asp_hist_si",
                "create attribute if not exists flash.prod_margin_bklg_st",
                "create attribute if not exists pii.person",

                "DROP ROLE IF EXISTS %s" % (role),
                "CREATE ROLE %s WITH GROUPS %s" % (role, testuser),

                """GRANT SELECT ON DATABASE `%s`
                HAVING ATTRIBUTE IN (`flash`.`cogs_bklg_cra`, `flash`.`cogs_bklg_si`, `flash`.`cogs_bklg_st`, `flash`.`cogs_hist_cra`,  `flash`.`cogs_hist_si`, `flash`.`cogs_hist_st`, `flash`.`qty_bklg_cra`, `flash`.`qty_bklg_si`, `flash`.`qty_bklg_st`, `flash`.`qty_hist_cra`, `flash`.`qty_hist_si`, `flash`.`qty_hist_st`, `flash`.`revenue_bklg_cra`, `flash`.`revenue_bklg_si`, `flash`.`revenue_bklg_st`,`flash`.`resale_bklg_st`, `flash`.`revenue_hist_cra`, `flash`.`revenue_hist_si`, `flash`.`revenue_hist_st`, `flash`.`resale_hist_st`, `flash`.`prod_margin_bklg_st`, `flash`.`asp_hist_st`, `flash`.`asp_hist_si`, `flash`.`revenue_cover_st`)
                WHERE ((in_set('ALL', user_attribute('customer_tier')) OR in_set(customer_tier, user_attribute('customer_tier')))
                AND (in_set('ALL', user_attribute('external_markets')) OR in_set(external_markets, user_attribute('external_markets')))
                AND (in_set('ALL', user_attribute('external_markets_alias')) OR in_set(external_markets_alias, user_attribute('external_markets_alias')))
                AND (in_set('ALL', user_attribute('internal_reporting_markets')) OR in_set(internal_reporting_markets, user_attribute('internal_reporting_markets')))
                AND (in_set('ALL', user_attribute('primary_end_markets')) OR in_set(primary_end_markets, user_attribute('primary_end_markets')))
                -- AND (in_set('ALL', user_attribute('prim_sec_end_markets')) OR in_set(primary_secondary_combo, user_attribute('prim_sec_end_markets')))
                AND (in_set('ALL', user_attribute('sub_seg_level1')) OR in_set(sub_seg_level1, user_attribute('sub_seg_level1')))
                AND (in_set('ALL', user_attribute('channel_geography')) OR in_set(channel_geography, user_attribute('channel_geography')))
                AND (in_set('ALL', user_attribute('channel_hier3')) OR in_set(channel_hier3, user_attribute('channel_hier3')))
                AND (in_set('ALL', user_attribute('customer_xu')) OR in_set(customer_xu, user_attribute('customer_xu')))
                AND (in_set('ALL', user_attribute('rep_name')) OR in_set(rep_name, user_attribute('rep_name')))
                AND (in_set('ALL', user_attribute('territory')) OR in_set(territory, user_attribute('territory')))
                -- AND (in_set('ALL', user_attribute('sector')) OR in_set(sector, user_attribute('sector')))
                -- AND (in_set('ALL', user_attribute('area')) OR in_set(area, user_attribute('area')))
                -- AND (in_set('ALL', user_attribute('internal_product_line')) OR in_set(internal_product_line, user_attribute('internal_product_line')))
                )
                TO ROLE `%s`""" % (db, role),

                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE fiscal_quarter  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE fiscal_quarter_org  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE fiscal_qtr_std  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE year_quarter    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE year_quarter_week   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtr_alias   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE week_alias  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE year_alias  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE fiscal_year flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE channel_distribution_type   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE channel_direct_type flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE channel_hier3   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE channel_geography   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE channel_geo_hier2   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE channel_distributor_name    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE part    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE extnd_product_dppg  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE extnd_product_device    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE extnd_product_family    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE extnd_product_sub_group flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE extnd_product_sub_group_alias   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE extnd_product_group flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE extnd_product_group_alias   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE extended_product_header flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE geometry    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE internal_product_dppg   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE internal_product_group  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE internal_product_family flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE internal_product_line   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE internal_product_division   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE internal_product_roll_up    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE part_external   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE part_pkg_pin    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE part_grade  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE part_speed  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE part_logic_cells    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE part_node   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE primary_end_markets flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE primary_secondary_combo flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE secondary_end_markets   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE external_markets    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE external_markets_alias  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE ext_mkt_level1  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE primary_end_market_alias    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE internal_reporting_markets  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE int_mkt_level1  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE sub_seg_level1  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE sub_segments    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE territory   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE sector  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE area    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE region  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE rep_name    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE rep_alias   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE default_territory   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE default_sector  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE default_area    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE default_region  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE default_rep_name    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE default_rep_alias   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE customer_tier   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE customer_tier_detail    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE customer_xu flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE customer_gu flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE hub_party   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE ship_customer_tier  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE ship_customer_tier_detail   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE ship_customer_xu    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE ship_customer_gu    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE ship_hub_party  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE scenario    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE record_type flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE demand_creation_category    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE demand_creation_category_alias  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE demand_creation_type    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE business_type   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_curr_qtr_backlog flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_curr_qtr_backlog flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_curr_qtr_backlog flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_actual_delinquency flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_curr_qtr_backlog  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_curr_qtr_backlog  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_curr_qtr_backlog  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_allocated_delinquency  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_curr_qtr_backlog  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_curr_qtr_backlog  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_curr_qtr_backlog  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_csd    flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_curr_qtr_delinquency  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_cfd    flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtr_plus2_bklg_rev  flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_revenue  flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_revenue  flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_revenue  flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_revenue_adjustments flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_revenue_adjustments flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_revenue_adjustments flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_revenue   flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_revenue   flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_revenue   flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_revenue   flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_revenue   flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_revenue   flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_revenue flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_revenue flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_curr_qtr_backlog_qty flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_curr_qtr_backlog_qty  flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_curr_qtr_backlog_qty flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_curr_qtr_backlog_qty  flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_curr_qtr_backlog_qty flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_curr_qtr_backlog_qty  flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_actual_delinquency_qty flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_allocated_delinquency_qty  flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_curr_qtr_backlog_qty  flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_curr_qtr_backlog_qty  flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_curr_qtr_backlog_qty  flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_csd_qty    flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_curr_qtr_delinquency_qty  flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_qty    flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_curr_qtr_backlog_cogs    flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_curr_qtr_backlog_cogs flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_curr_qtr_backlog_cogs    flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_curr_qtr_backlog_cogs flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_curr_qtr_backlog_cogs    flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_curr_qtr_backlog_cogs flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_actual_delinquency_cogs    flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_allocated_delinquency_cogs flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_curr_qtr_backlog_cogs flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_curr_qtr_backlog_cogs flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_curr_qtr_backlog_cogs flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_csd_cogs   flash.cogs_bklg_st flash.measure pii.person" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_curr_qtr_delinquency_cogs flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_cogs   flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_qty  flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_qty   flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_adjustments_qty flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_qty  flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_qty   flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_adjustments_qty flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_qty  flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_qty   flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_adjustments_qty flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_qty   flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_qty   flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_qty   flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_qty flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_qty flash.measure flash.qty_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_cogs flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_cogs  flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_adjustments_cogs    flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_cogs flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_cogs  flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_adjustments_cogs    flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_cogs flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_cogs  flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_adjustments_cogs    flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_cogs  flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_cogs  flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_cogs  flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_cogs    flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_cogs    flash.cogs_hist_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtr_plus2_bklg_qty  flash.measure flash.qty_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtr_plus2_bklg_cogs flash.cogs_bklg_st flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_curr_qtr_backlog_cra flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_curr_qtr_backlog_cra flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_curr_qtr_backlog_cra flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtr_plus2_bklg_rev_cra  flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_curr_qtr_backlog_cra  flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_curr_qtr_backlog_cra  flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_curr_qtr_backlog_cra  flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_curr_qtr_backlog_cra  flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_curr_qtr_backlog_cra  flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_curr_qtr_backlog_cra  flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_csd_cra    flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_cra    flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_revenue_cra  flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_revenue_cra  flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_revenue_cra  flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_revenue_adjustments_cra flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_revenue_adjustments_cra flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_revenue_adjustments_cra flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_revenue_cra   flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_revenue_cra   flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_revenue_cra   flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_revenue_cra flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_revenue_cra flash.measure flash.revenue_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_sell_in    flash.measure flash.revenue_bklg_si" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_revenue_sell_in flash.measure flash.revenue_hist_si" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_sell_in flash.measure flash.revenue_hist_si" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE actual_curr_qtr_backlog flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE allocated_curr_qtr_backlog  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_actual_and_allocation  flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE adjustment  flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_curr_qtr_backlog_qty_cra flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_curr_qtr_backlog_qty_cra flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_curr_qtr_backlog_qty_cra flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtr_plus2_bklg_qty_cra  flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_curr_qtr_backlog_qty_cra  flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_curr_qtr_backlog_qty_cra  flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_curr_qtr_backlog_qty_cra  flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_curr_qtr_backlog_qty_cra  flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_curr_qtr_backlog_qty_cra  flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_curr_qtr_backlog_qty_cra  flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_csd_qty_cra    flash.measure flash.qty_bklg_cra pii.person" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_qty_cra    flash.measure flash.qty_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_qty_cra  flash.measure flash.qty_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_qty_cra  flash.measure flash.qty_hist_cra pii.person" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_qty_cra  flash.measure flash.qty_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_adjustments_qty_cra flash.measure flash.qty_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_adjustments_qty_cra flash.measure flash.qty_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_adjustments_qty_cra flash.measure flash.qty_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_qty_cra   flash.measure flash.qty_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_qty_cra   flash.measure flash.qty_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_qty_cra   flash.measure flash.qty_hist_cra pii.person" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_qty_cra flash.measure flash.qty_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_qty_cra flash.measure flash.qty_hist_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_qty_sell_in    flash.measure flash.qty_bklg_si" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_qty_sell_in flash.measure flash.qty_hist_si" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_qty_sell_in flash.measure flash.qty_hist_si" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_curr_qtr_backlog_cogs_cra    flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_curr_qtr_backlog_cogs_cra    flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_curr_qtr_backlog_cogs_cra    flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtr_plus2_bklg_cogs_cra flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_curr_qtr_backlog_cogs_cra flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_curr_qtr_backlog_cogs_cra flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_curr_qtr_backlog_cogs_cra flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_curr_qtr_backlog_cogs_cra flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_curr_qtr_backlog_cogs_cra flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_curr_qtr_backlog_cogs_cra flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_csd_cogs_cra   flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_cogs_cra   flash.cogs_bklg_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_cogs_cra flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_cogs_cra flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_cogs_cra flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_adjustments_cogs_cra    flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_adjustments_cogs_cra    flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_adjustments_cogs_cra    flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_cogs_cra  flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_cogs_cra  flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_cogs_cra  flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_cogs_cra    flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_cogs_cra    flash.cogs_hist_cra flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_cogs_sell_in   flash.cogs_bklg_si flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_cogs_sell_in    flash.cogs_hist_si flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_cogs_sell_in    flash.cogs_hist_si flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_curr_qtr_backlog_resale  flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_curr_qtr_backlog_resale  flash.measure flash.resale_bklg_st pii.person" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_curr_qtr_backlog_resale  flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_actual_delinquency_resale  flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_curr_qtr_backlog_resale   flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_curr_qtr_backlog_resale   flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_curr_qtr_backlog_resale   flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_allocated_delinquency_resale   flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_curr_qtr_backlog_resale   flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_curr_qtr_backlog_resale   flash.measure flash.resale_bklg_st pii.person" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_curr_qtr_backlog_resale   flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_resale_csd flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_curr_qtr_delinquency_resale   flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE curr_qtr_backlog_resale flash.measure flash.resale_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtr_plus2_bklg_res  flash.measure" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_actual_resale   flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_actual_resale   flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_actual_resale   flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_resale_adjustments  flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_resale_adjustments  flash.measure flash.revenue_hist_st pii.person" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_resale_adjustments  flash.measure flash.revenue_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_1_allocated_resale    flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_2_allocated_resale    flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE month_3_allocated_resale    flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_1_resale    flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_2_resale    flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE total_month_3_resale    flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE qtd_resale  flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_resale  flash.measure flash.resale_hist_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_revenue_qtr_ago flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_revenue_cra_qtr_ago flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_sell_in_qtr_ago flash.measure flash.revenue_bklg_si" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_revenue_qtr_ago_turns   flash.measure flash.revenue_bklg_st" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_revenue_cra_qtr_ago_turns   flash.measure flash.revenue_bklg_cra" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flash_total_sell_in_qtr_ago_turns   flash.measure flash.revenue_bklg_si" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_time_key flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_part_key flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_geography_key    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_channel_key  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_cust_key_disti   flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_cust_key_end flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_city_key flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_cust_key_end_xcm flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_city_key_xcm flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_sales_hierarchy_key  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_mkt_seg_key  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_default_sales_hierarchy_key  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_cust_key_ship_xcm    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_business_type_key    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE dm_demand_creation_key  flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE flag_14_week    flash.dimension" % (db, tbl1),
                "ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE week_no flash.dimension" % (db, tbl1),
            ]

            print("Creating environment with DDLs")
            start = time.time()
            for ddl in ddls:
                conn.execute_ddl(ddl)
            end = time.time()
            print("Done creating environment with DDLs: %s seconds" % (end-start))

        # We put the query on a separate connection as it makes debugging the logs
        # a bit easier.
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.enable_token_auth(token_str=testuser)

            print("Executing query")
            start = time.time()
            conn.scan_as_json('select * from %s.%s' % (db, tbl1))
            end = time.time()
            print("Done query: %s seconds" % (end-start))


# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for auth in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=no-else-return
# pylint: disable=too-many-lines


import datetime
import json
import os
import warnings

import pytest
import numpy

import requests

from okera._thrift_api import TRecordServiceException
from okera.tests import pycerebro_test_common as common

API_URL = "http://%s:%s" % \
    (os.environ['ODAS_TEST_HOST'], os.environ['ODAS_TEST_PORT_REST_SERVER'])

def get_scan_as_json(conn, dataset):
    data = conn.scan_as_json(
        dataset, strings_as_utf8=True,
        max_records=10, max_client_process_count=1)
    return data

def get_scan_as_pandas(conn, dataset):
    def convert_types(datum):
        if isinstance(datum, datetime.datetime):
            return datum.replace(tzinfo=None).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        elif isinstance(datum, numpy.int64):
            return int(datum)
        elif isinstance(datum, dict):
            for key, value in datum.items():
                datum[key] = convert_types(value)
            return datum
        elif isinstance(datum, list):
            return [convert_types(child) for child in datum]

        return datum

    data = conn.scan_as_pandas(
        dataset, strings_as_utf8=True,
        max_records=10, max_client_process_count=1)
    data = data.replace({numpy.nan: None})
    data = data.to_dict('records')
    return convert_types(data)

def get_scan_from_rest(dataset):
    headers = {'content-type': 'application/json'}
    query = {'query': 'SELECT * FROM %s' % dataset}
    response = requests.post(
        API_URL + '/api/scan?records=10',
        json=query, headers=headers)
    result = json.loads(response.text)
    return result

def format(data):
    return json.dumps(data, sort_keys=True, indent=1)

@pytest.mark.parametrize("dataset", [
    'rs_complex.array_struct_array',
    'rs_complex.array_struct_t',
    'rs_complex.array_t',
    'rs_complex.avrotbl',
    'rs_complex.bytes_type',
    'rs_complex.bytes_type_file',
    'rs_complex.enum_type',
    'rs_complex.enum_type_default',
    'rs_complex.map_t',
    'rs_complex.market_decide_single_avro',
    'rs_complex.market_v20_single',
    'rs_complex.market_v30_single',
    'rs_complex.multiple_structs_nested',
    'rs_complex.strarray_t',
    'rs_complex.strarray_t_view',
    'rs_complex.struct_array_struct',
    'rs_complex.struct_nested',
    'rs_complex.struct_nested_s1',
    'rs_complex.struct_nested_s1_f1',
    'rs_complex.struct_nested_s1_s2',
    'rs_complex.struct_nested_view',
    'rs_complex.struct_t',
    'rs_complex.struct_t2',
    'rs_complex.struct_t3',
    'rs_complex.struct_t_id',
    'rs_complex.struct_t_s1',
    'rs_complex.struct_t_view',
    'rs_complex.struct_t_view2',
    'rs_complex.user_phone_numbers',
    'rs_complex.user_phone_numbers_map',
    'rs_complex.users',
    'rs_complex.view_over_multiple_structs',
    'rs_complex_parquet.array_struct_array',
    'rs_complex_parquet.array_struct_map_t',
    'rs_complex_parquet.array_struct_t',
    'rs_complex_parquet.array_struct_t2',
    'rs_complex_parquet.array_t',
    'rs_complex_parquet.bytes_type',
    'rs_complex_parquet.enum_type',
    'rs_complex_parquet.map_array',
    'rs_complex_parquet.map_array_t2',
    'rs_complex_parquet.map_struct_array_t',
    'rs_complex_parquet.map_struct_array_t_view',
    'rs_complex_parquet.map_struct_t',
    'rs_complex_parquet.map_struct_t_view',
    'rs_complex_parquet.map_t',
    'rs_complex_parquet.strarray_t',
    'rs_complex_parquet.struct_array_struct',
    'rs_complex_parquet.struct_nested',
    'rs_complex_parquet.struct_nested_s1',
    'rs_complex_parquet.struct_nested_s1_f1',
    'rs_complex_parquet.struct_nested_s1_s2',
    'rs_complex_parquet.struct_nested_view',
    'rs_complex_parquet.struct_t',
    'rs_complex_parquet.struct_t2',
    'rs_complex_parquet.struct_t3',
    'rs_complex_parquet.struct_t_id',
    'rs_complex_parquet.struct_t_s1',
    'rs_complex_parquet.struct_t_view',
    'rs_complex_parquet.struct_t_view2',
    # # The ones below don't work because the /scan API returns the wrong value
    # # for decimals
    # 'rs_complex_parquet.spark_all_mixed_compression',
    # 'rs_complex_parquet.spark_gzip',
    # 'rs_complex_parquet.spark_snappy',
    # 'rs_complex_parquet.spark_snappy_part',
    # 'rs_complex_parquet.spark_uncompressed',
    # 'rs_complex_parquet.spark_uncompressed_legacy_format',
    'customer.zd277_complex',
    'rs_json_format.complex_c1_case_sensitive',
    'rs_json_format.complex_c1_usecase',
    'rs_json_format.complex_nike_usecase',
    # NOT WORKING array of arrays is not allowed, scan error
    # 'rs_json_format.json_array_arrays',
    'rs_json_format.json_array_struct',
    'rs_json_format.json_arrays_test',
    # # These tables don't work because Pandas and JSON don't serialize to the
    # # exact same types (e.g. float vs int), even though the overall data is correct)
    # 'rs_json_format.json_inferred',
    # 'rs_json_format.json_primitives',
    # 'rs_json_format.json_primitives_inferred',
    'rs_json_format.json_struct',
    'rs_json_format.json_struct_array',
    'rs_json_format.json_struct_nested',
    'rs_json_format.primitives_with_array',
])

def test_basic(dataset):
    print("Testing: " + dataset)
    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    ctx = common.get_test_context()
    with common.get_planner(ctx) as conn:
        json_data = get_scan_as_json(conn, dataset)
        pandas_data = get_scan_as_pandas(conn, dataset)
        rest_data = get_scan_from_rest(dataset)
        assert format(json_data) == format(pandas_data)
        assert format(json_data) == format(rest_data)

def test_pandas_empty_result():
    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    ctx = common.get_test_context()
    with common.get_planner(ctx) as conn:
        pandas_data = conn.scan_as_pandas('select s1 from rs_complex.struct_t where 1=0')
        columns = list(pandas_data.columns)
        assert columns == ['s1']

        pandas_data = conn.scan_as_pandas(
            'select a1 from rs_complex.array_struct_array where 1=0')
        columns = list(pandas_data.columns)
        assert columns == ['a1']

        pandas_data = conn.scan_as_pandas(
            'select s1, s1.a1 as x from rs_complex.struct_array_struct where 1=0')
        columns = list(pandas_data.columns)
        assert columns == ['s1', 'x']

        pandas_data = conn.scan_as_pandas(
            'select * from rs_complex_parquet.map_struct_array_t where 1=0')
        columns = list(pandas_data.columns)
        assert columns == ['id', 'str_arr', 'struct_map']

def test_json_nonutf8():
    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    ctx = common.get_test_context()
    with common.get_planner(ctx) as conn:
        data = conn.scan_as_json(
            'select * from rs.encoding',
            strings_as_utf8=False)
        assert data == [{'uid': b'ABC123', 'message': b'\xe8 bene'}]

def test_null_map():
    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    ctx = common.get_test_context()
    with common.get_planner(ctx) as conn:
        json_data = get_scan_as_json(conn, 'fastparquet.map_array_parq')
        pandas_data = get_scan_as_pandas(conn, 'fastparquet.map_array_parq')
        assert format(json_data) == format(pandas_data)

def assert_scan_output(qry, expected, is_scan=True, as_utf8=False, limit=None):
    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    ctx = common.get_test_context()
    qry = qry.rstrip()
    with common.get_planner(ctx) as conn:
        print("Running query: %s" % qry)
        if is_scan:
            data = conn.scan_as_json(qry, strings_as_utf8=as_utf8, max_records=limit)
        else:
            data = conn.execute_ddl(qry)
        print("data is: ")
        print(data)
        #print(json.dumps(data, indent=2))
        print("expected is: ")
        print(expected)
        #print(json.dumps(expected, indent=2))
        assert data == expected

def assert_query(sql, expected, skip_inline_view=False, skip_count=False,
                 skip_catalog_view=False):
    assert_scan_output(sql, expected, as_utf8=True)
    if not skip_inline_view:
        wrapper_view = "SELECT * FROM (%s) do_not_use_name" % sql
        assert_scan_output(wrapper_view, expected, as_utf8=True)
    if not skip_count:
        count_view = "SELECT count(*) as cnt FROM (%s) do_not_use_name" % sql
        assert_scan_output(count_view, [{'cnt': len(expected)}], as_utf8=True)
    if not skip_catalog_view:
        with common.get_planner(common.get_test_context()) as conn:
            with common.TmpView(conn, sql) as view:
                assert_scan_output('SELECT * FROM %s' % view.name(),
                                   expected, as_utf8=True)
                count_view = 'SELECT count(*) as cnt FROM %s' % view.name()
                assert_scan_output(count_view, [{'cnt': len(expected)}], as_utf8=True)

def assert_scan_exception(qry, expected, is_scan=True):
    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    ctx = common.get_test_context()
    with common.get_planner(ctx) as conn:
        try:
            if is_scan:
                result = conn.scan_as_json(qry, strings_as_utf8=False)
            else:
                result = conn.execute_ddl(qry)
            assert result
        except TRecordServiceException as ex:
            assert expected in str(ex.detail)

# pylint: disable=bad-continuation
def test_unnest():
    assert_query('''
        select id, str_arr.item
        from rs_complex.strarray_t, rs_complex.strarray_t.str_arr
        ''',
        [{'id': 456, 'item': 'a'}, {'id': 456, 'item': 'b'},
         {'id': 457, 'item': 'cde'}, {'id': 457, 'item': ''},
         {'id': 458, 'item': 'fg'},
         {'id': 458, 'item': 'ijlk'}])

    assert_query('''
        select id, str_arr.item item
        from rs_complex.strarray_t, rs_complex.strarray_t.str_arr where id = 456
        ''',
        [{'id': 456, 'item': 'a'}, {'id': 456, 'item': 'b'}])

    # TODO(bug): Illegal reference to non-materialized tuple
    assert_query('''
        select a1.item, mask(a1.item.f1)
        from rs_complex.array_struct_t, rs_complex.array_struct_t.a1
        ''',
        [{'item': {'f2': 'c', 'f1': 'ab'}, 'mask(a1': {'item': {'f1)': 'XX'}}},
         {'item': {'f2': '', 'f1': 'def'}, 'mask(a1': {'item': {'f1)': 'XXX'}}},
         {'item': {'f2': 'ij', 'f1': 'g'}, 'mask(a1': {'item': {'f1)': 'X'}}}],
        skip_inline_view=True,
        skip_catalog_view=True)

    # TODO(usability): Error creating view: Invalid column/field name: item.f2
    assert_query('''
        select
            a1.item.f1 as raw_f1,
            concat(a1.item.f1, '_f1') as concat_f1,
            a1.item.f2,
            concat(a1.item.f2, '_f2') as concat_f2
        from rs_complex.array_struct_t, rs_complex.array_struct_t.a1
        ''',
        [{'raw_f1': 'ab', 'item': {'f2': 'c'},
          'concat_f1': 'ab_f1', 'concat_f2': 'c_f2'},
         {'raw_f1': 'def', 'item': {'f2': ''},
          'concat_f1': 'def_f1', 'concat_f2': '_f2'},
         {'raw_f1': 'g', 'item': {'f2': 'ij'},
          'concat_f1': 'g_f1', 'concat_f2': 'ij_f2'}],
        skip_catalog_view=True)

    # TODO(usability): Error creating view: Invalid column/field name: item.f2
    assert_query('''
        select a1.item.f2, mask(a1.item.f1)
        from rs_complex.array_struct_t, rs_complex.array_struct_t.a1
        ''',
        [{'item': {'f2': 'c'}, 'mask(a1': {'item': {'f1)': 'XX'}}},
         {'item': {'f2': ''}, 'mask(a1': {'item': {'f1)': 'XXX'}}},
         {'item': {'f2': 'ij'}, 'mask(a1': {'item': {'f1)': 'X'}}}],
        skip_catalog_view=True)

    assert_query('''
        select id, str_arr.item
        from rs_complex.strarray_t, rs_complex.strarray_t.str_arr
        where str_arr.item = 'cde'
        ''',
        [{'id': 457, 'item': 'cde'}])

    assert_scan_exception('''
        select tokenize(partyroles.item.partykey) as tokenized_key
        from chase.zd1238_4.partyroles
        ''',
        "AnalysisException: Correlated table 'chase.zd1238_4' must " \
        "be specified before the collection reference " \
        "'chase.zd1238_4.partyroles' in the from clause")

    # TODO: Make execddl return the raw output for explain plan, then enable this.
    # assert_scan_output(
    #     'EXPLAIN SELECT subscription_key_tokenized FROM chase.subscription_party_view',
    #     '00:SCAN HDFS [chase.zd1238_4] '\
    #     '    partitions=1/1 files=1 size=54.05KB '\
    #     '    predicates: !empty(chase.zd1238_4.partyroles)',
    #     False
    # )

    assert_query('''
        select a1.* from rs_complex.array_struct_t, rs_complex.array_struct_t.a1
        ''',
        [{'f2': 'c', 'f1': 'ab'},
         {'f2': '', 'f1': 'def'},
         {'f2': 'ij', 'f1': 'g'}])

    # TODO(bug): (inline and catalog view): Illegal reference to non-materialized tuple
    assert_query('''
        select a1.item.* from rs_complex.array_struct_t, rs_complex.array_struct_t.a1
        ''',
        [{'item': {'f2': 'c', 'f1': 'ab'}},
         {'item': {'f2': '', 'f1': 'def'}},
         {'item': {'f2': 'ij', 'f1': 'g'}}],
        skip_inline_view=True,
        skip_catalog_view=True)

    assert_query('''
        select a1, mask(a1.item.f1) as masked
        from rs_complex.array_struct_t, rs_complex.array_struct_t.a1
        ''',
        [{'a1': [{'f1': 'ab', 'f2': 'c'}], 'masked': 'XX'},
         {'a1': [{'f1': 'def', 'f2': ''}], 'masked': 'XXX'},
         {'a1': [{'f1': 'g', 'f2': 'ij'}], 'masked': 'X'}])

    assert_query('''
        select a1.* from rs_complex.array_struct_t, rs_complex.array_struct_t.a1
        ''',
        [{'f1': 'ab', 'f2': 'c'},
         {'f1': 'def', 'f2': ''},
         {'f1': 'g', 'f2': 'ij'}])

    # TODO(usability): Error creating view: Invalid column/field name: item.f2
    assert_query('''
        select a1.item.a2
        from rs_complex.array_struct_array, rs_complex.array_struct_array.a1
        ''',
        [{'item': {'a2': ['jk']}},
         {'item': {'a2': ['l']}},
         {'item': {'a2': ['']}}],
        skip_catalog_view=True)

    # TODO(bug): (inline and catalog view): Illegal reference to non-materialized tuple
    assert_query('''
        select a1.item
        from rs_complex.array_struct_array, rs_complex.array_struct_array.a1
        ''',
        [{'item': {'f1': 'ab', 'f2': 'c', 'a2': ['jk']}},
         {'item': {'f1': 'def', 'f2': '', 'a2': ['l']}},
         {'item': {'f1': 'g', 'f2': 'ij', 'a2': ['']}}],
        skip_inline_view=True,
        skip_catalog_view=True)

    assert_query('''
        select a1.item.f2 as f2, mask(a1.item.f1) as masked
        from rs_complex.array_struct_t, rs_complex.array_struct_t.a1
        ''',
        [{'f2': 'c', 'masked': 'XX'},
         {'f2': '', 'masked': 'XXX'},
         {'f2': 'ij', 'masked': 'X'}])

    assert_query('''
        select a1.item.f2 as f2, mask(a1.item.f1) as masked
        from rs_complex.array_struct_t, rs_complex.array_struct_t.a1
        where a1.item.f2 = 'c'
        ''',
        [{'f2': 'c', 'masked': 'XX'}])

    # TODO(bug): (inline and catalog view): Illegal reference to non-materialized tuple
    assert_query('''
        select a.s1 from rs_complex.struct_array_struct a, a.s1.a1
        ''',
        [{'s1': {'a1': [{'f1': 'ab', 'f2': 'c', 'a2': ['jk']},
                        {'f1': 'def', 'f2': '', 'a2': ['l']}]}},
         {'s1': {'a1': [{'f1': 'ab', 'f2': 'c', 'a2': ['jk']},
                        {'f1': 'def', 'f2': '', 'a2': ['l']}]}},
         {'s1': {'a1': [{'f1': 'g', 'f2': 'ij', 'a2': ['']}]}}],
        skip_inline_view=True,
        skip_catalog_view=True)

    # TODO(bug): (inline and catalog view): Illegal reference to non-materialized tuple
    assert_query('''
        select a.s1 from rs_complex.struct_array_struct a, a.s1.a1
        where a1.item.f1 = 'ab'
        ''',
        [{'s1': {'a1': [{'f1': 'ab', 'f2': 'c', 'a2': ['jk']},
                        {'f1': 'def', 'f2': '', 'a2': ['l']}]}}],
        skip_inline_view=True,
        skip_catalog_view=True)

    # FIXME(bug): Inline/catalog view returning empty
    assert_query('''
        select a.s1, a1.* from rs_complex.struct_array_struct a, a.s1.a1
        ''',
        [{'s1': {'a1': [{'f2': 'c', 'a2': ['jk'], 'f1': 'ab'},
                        {'f2': '', 'a2': ['l'], 'f1': 'def'}]},
          'f2': 'c', 'a2': ['jk'], 'f1': 'ab'},
         {'s1': {'a1': [{'f2': 'c', 'a2': ['jk'], 'f1': 'ab'},
                        {'f2': '', 'a2': ['l'], 'f1': 'def'}]},
          'f2': '', 'a2': ['l'], 'f1': 'def'},
         {'s1': {'a1': [{'f2': 'ij', 'a2': [''], 'f1': 'g'}]},
          'f2': 'ij', 'a2': [''], 'f1': 'g'}],
        skip_inline_view=True,
        skip_catalog_view=True)

    # FIXME(bug): Inline/catalog view returning empty
    assert_query('''
        select a.s1, a1.*
        from rs_complex.struct_array_struct a, a.s1.a1
        where a1.item.f1 = 'ab'
        ''',
        [{'f2': 'c', 'f1': 'ab', 'a2': ['jk'],
          's1': {'a1': [{'f2': 'c', 'f1': 'ab', 'a2': ['jk']},
                        {'f2': '', 'f1': 'def', 'a2': ['l']}]}}],
        skip_inline_view=True,
        skip_catalog_view=True)

    ## Test on many Array types in same table.
    assert_query('''
        select int32, int_arr, str_arr from rs_complex_parquet.spark_snappy
        ''',
        [{'int_arr': [1, 2], 'int32': 1, 'str_arr': ['t1', 't2']},
         {'int_arr': [3, 4], 'int32': 2, 'str_arr': ['t3', 't4']}])

    assert_query('''
        select int32, str_arr.item, int_arr
        from rs_complex_parquet.spark_gzip, rs_complex_parquet.spark_gzip.str_arr
        ''',
        [{'item': 't1', 'int_arr': [1, 2], 'int32': 1},
         {'item': 't2', 'int_arr': [1, 2], 'int32': 1},
         {'item': 't3', 'int_arr': [3, 4], 'int32': 2},
         {'item': 't4', 'int_arr': [3, 4], 'int32': 2}])

    assert_query('''
        select int32,
            int_arr.item as i,
            str_arr.item as s,
            mask(str_arr.item) as str_masked
        from
            rs_complex_parquet.spark_snappy,
            rs_complex_parquet.spark_snappy.str_arr,
            rs_complex_parquet.spark_snappy.int_arr
        ''',
        [{'i': 1, 's': 't1', 'str_masked': 'XX', 'int32': 1},
         {'i': 2, 's': 't1', 'str_masked': 'XX', 'int32': 1},
         {'i': 1, 's': 't2', 'str_masked': 'XX', 'int32': 1},
         {'i': 2, 's': 't2', 'str_masked': 'XX', 'int32': 1},
         {'i': 3, 's': 't3', 'str_masked': 'XX', 'int32': 2},
         {'i': 4, 's': 't3', 'str_masked': 'XX', 'int32': 2},
         {'i': 3, 's': 't4', 'str_masked': 'XX', 'int32': 2},
         {'i': 4, 's': 't4', 'str_masked': 'XX', 'int32': 2}])

    assert_query('''
        select
            int32,
            inner_str_arr.item as srt_arr_str,
            mask(inner_str_arr.item) as masked_srt_arr_str,
            mask(struct_arr.item.inner_str) masked_struct_str
        from
            rs_complex_parquet.spark_snappy,
            rs_complex_parquet.spark_snappy.struct_t.inner_str_arr,
            rs_complex_parquet.spark_snappy.struct_arr
        ''',
        [{'masked_srt_arr_str': 'XXX', 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'srt_arr_str': 'In1'},
         {'masked_srt_arr_str': 'XXX', 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'srt_arr_str': 'In1'},
         {'masked_srt_arr_str': 'XXX', 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'srt_arr_str': 'In2'},
         {'masked_srt_arr_str': 'XXX', 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'srt_arr_str': 'In2'},
         {'masked_srt_arr_str': 'XXX', 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'srt_arr_str': 'In7'},
         {'masked_srt_arr_str': 'XXX', 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'srt_arr_str': 'In7'},
         {'masked_srt_arr_str': 'XXX', 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'srt_arr_str': 'In8'},
         {'masked_srt_arr_str': 'XXX', 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'srt_arr_str': 'In8'}])

    ## One test with array of int, string and struct<string>.
    ## we may not write many such tests as its too verbose output but this should
    ## be good coverage for flattened output.
    assert_query('''
        select
            int32,
            int_arr.item as flat_int,
            mask(str_arr.item) as masked_str,
            struct_arr.item.inner_str as raw_struct_str,
            mask(struct_arr.item.inner_str) masked_struct_str
        from
            rs_complex_parquet.spark_snappy,
            rs_complex_parquet.spark_snappy.str_arr,
            rs_complex_parquet.spark_snappy.int_arr,
            rs_complex_parquet.spark_snappy.struct_arr
        ''',
        [{'masked_str': 'XX', 'flat_int': 1, 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr2'},
         {'masked_str': 'XX', 'flat_int': 1, 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr3'},
         {'masked_str': 'XX', 'flat_int': 2, 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr2'},
         {'masked_str': 'XX', 'flat_int': 2, 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr3'},
         {'masked_str': 'XX', 'flat_int': 1, 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr2'},
         {'masked_str': 'XX', 'flat_int': 1, 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr3'},
         {'masked_str': 'XX', 'flat_int': 2, 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr2'},
         {'masked_str': 'XX', 'flat_int': 2, 'int32': 1,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr3'},
         {'masked_str': 'XX', 'flat_int': 3, 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr5'},
         {'masked_str': 'XX', 'flat_int': 3, 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr6'},
         {'masked_str': 'XX', 'flat_int': 4, 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr5'},
         {'masked_str': 'XX', 'flat_int': 4, 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr6'},
         {'masked_str': 'XX', 'flat_int': 3, 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr5'},
         {'masked_str': 'XX', 'flat_int': 3, 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr6'},
         {'masked_str': 'XX', 'flat_int': 4, 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr5'},
         {'masked_str': 'XX', 'flat_int': 4, 'int32': 2,
          'masked_struct_str': 'XXXXXXXXX', 'raw_struct_str': 'InnerStr6'}])

    assert_query('''
        select id, str_arr.item item
        from rs_complex.strarray_t, rs_complex.strarray_t.str_arr where id = 456
        ''',
        [{'item': 'a', 'id': 456}, {'item': 'b', 'id': 456}])

    assert_scan_exception('''
        select id, str_arr.item item
        from rs_complex.strarray_t.str_arr, rs_complex.strarray_t where id = 456
        ''',
        "Correlated table 'rs_complex.strarray_t' must be specified before " \
        "the collection reference 'rs_complex.strarray_t.str_arr' in the from clause ")

    assert_query('''
        select int32, int_arr.item
        from rs_complex_parquet.spark_gzip, rs_complex_parquet.spark_gzip.int_arr
        ''',
        [{'int32': 1, 'item': 1}, {'int32': 1, 'item': 2},
         {'int32': 2, 'item': 3}, {'int32': 2, 'item': 4}]
    )

    assert_query('''
        select int32, int_arr.item
        from rs_complex_parquet.spark_gzip, rs_complex_parquet.spark_gzip.int_arr
        ''',
        [{'int32': 1, 'item': 1}, {'int32': 1, 'item': 2},
         {'int32': 2, 'item': 3}, {'int32': 2, 'item': 4}]
    )

    assert_query('''
        select * from rs_complex.pn_view
        ''',
        [{'uid': 123, 'item': '111-222-3333'}, {'uid': 123, 'item': '444-555-6666'},
         {'uid': 234, 'item': '222-333-4444'}, {'uid': 234, 'item': '555-666-7777'},
         {'uid': 345, 'item': '111-222-5555'}]
    )

    assert_query('''
        select count(*) as unnest_view_count from rs_complex.pn_view
        ''',
        [{'unnest_view_count': 5}]
    )

    assert_query('''
        select count(*) as unnest_view_count from rs_complex.pn_view
        where uid = 234
        ''',
        [{'unnest_view_count': 2}]
    )

    ## Some GROUP BY/HAVING tests
    assert_query('''
        select count(item) as item_count, uid from rs_complex.pn_view
        GROUP BY uid HAVING count(item) > 1
        ''',
        [{'item_count': 2, 'uid': 123}, {'item_count': 2, 'uid': 234}]
    )
    assert_query('''
        select count(item) as item_count, uid from rs_complex.pn_view
        GROUP BY uid HAVING count(item) >= 1'''
        ,
        [{'item_count': 1, 'uid': 345}, {'item_count': 2, 'uid': 123},
         {'item_count': 2, 'uid': 234}]
    )
    assert_query('''
        select count(item) as item_count, uid from rs_complex.pn_view
        GROUP BY uid HAVING count(item) = 1
        ''',
        [{'item_count': 1, 'uid': 345}]
    )

def test_array_of_arrays():
    assert_scan_output(
        'select * from rs_json_format.json_array_arrays',
        [{'arr_outer': [[1, 2], [1200000, 345600]], 'arr_string': [b'a', b'b'],
          'str_val': b'abc', 'int_val': 1},
         {'arr_outer': [[3, 4], [1200000, 345600]], 'arr_string': [b'c', b'd'],
          'str_val': b'def', 'int_val': 4}])

    assert_scan_output(
        'select * from parquet_testing.nested_lists_snappy',
        [{'a': [[[b'a', b'b'], [b'c']], [None, [b'd']]], 'b': 1},
         {'a': [[[b'a', b'b'], [b'c', b'd']], [None, [b'e']]], 'b': 1},
         {'a': [[[b'a', b'b'], [b'c', b'd'], [b'e']], [None, [b'f']]], 'b': 1}])

    assert_scan_output(
        'select payload.shas from rs_json_format.gharch_test_data ' +
        'where id = \'1201654257\'',
        [
            {'payload':
                {'shas':[
                        [b'4d1085635d642ba60071341dd46d38216bff1313',
                        b'86d2a548fee49abc43bd1c251300d7089748195b@mitechie.com',
                        b'Garden it up',
                        b'Richard Harding'
                        ]
                    ]
                }
            }
        ]
    )

def test_full_unnest_with_long_column_names():
    ctx = common.get_test_context()
    with common.get_planner(ctx) as conn:
        plan = conn.plan('select ** from `rs_json_format`.`deeply_nested_json`')
        schema_cols = plan.schema.cols

        assert len(schema_cols) == 4
        assert schema_cols[0].name == 'ashort'
        assert schema_cols[1].name == 'a12__b12__c12__d12__e12__f12__g12__h12__i12__j12__k12__l1234567890__m1234567890__n1234567890__o1234567890__p1234567890__foo'
        assert schema_cols[2].name == 'a12__b12__c12__d12__e12__f12__g12__h12__i12__j12__k12__l12__m1234567890__n1234567890__o1234567890__p1234567890__baz__bar'
        assert schema_cols[3].name == 'a12__b12__c12__d12__e12__f12__g12__h12__i12__j12__k12__l12__m1234567890__n1234567890__array_o1234567890__p1234567890__baz__bar'


def test_full_unnest():
    def validate(t, expected):
        assert_scan_output('select ** from %s' % t, expected, as_utf8=True, limit=1)

    # Array types
    validate('rs_complex.array_t', [{'int_arr': 1, 'id': 456}])
    validate('rs_complex.array_struct_t', [{'a1__f2': 'c', 'a1__f1': 'ab'}])
    validate('rs_complex.strarray_t', [{'str_arr': 'a', 'id': 456}])
    validate('rs_complex.array_struct_array',
             [{'a1__a2': 'jk', 'a1__f1': 'ab', 'a1__f2': 'c'}])
    validate('rs_complex.struct_array_struct',
             [{'s1__a1__a2': 'jk', 'id': 100, 's1__a1__f2': 'c',
               's1__a1__f1': 'ab'}])
    validate('rs_complex.multiple_structs_nested',
             [{'struct1__array1__subfield11': 'namename22',
               'struct1__array1__subfield12': 'valuevalue22',
               'struct1__array2__subfield21': 'code5787878'}])
    validate('rs_complex.user_phone_numbers',
             [{'phone_numbers__uid': 123,
               'phone_numbers__numbers': '111-222-3333',
               'phone_numbers__user': 'john'}])
    validate('rs_complex.avrotbl',
             [{'messages__incident__escalation_policy__type':
                  'escalation_policy_reference',
               'messages__incident__service__acknowledgement_timeout': None,
               'messages__incident__escalation_policy__self': 'S2R6',
               'messages__incident__alert_counts__triggered': 1,
               'messages__incident__first_trigger_log_entry__self': 'BHVZH38',
               'messages__incident__assignments__assignee__html_url': '',
               'messages__incident__alert_counts__all': 1,
               'messages__incident__assignments__assignee__id': 'PJ7CR',
               'messages__incident__service__integrations__type':
                  'event_transformer_api_inbound_integration_reference',
               'messages__webhook__webhook_object__html_url': 'P7RAELC',
               'messages__incident__service__name': 'Card',
               'messages__incident__service__teams__type': 'team_reference',
               'messages__incident__service__escalation_policy__id': 'R6',
               'messages__incident__basic_alert_grouping': None,
               'messages__incident__service__summary': 'Card',
               'messages__incident__last_status_change_by__type': 'service_reference',
               'messages__log_entries__id': 'R3OWU2PBO5L90DLF76BS0LTTEV',
               'messages__incident__teams__html_url': '59ZN5',
               'messages__incident__pending_actions': None,
               'messages__incident__service__integrations__summary': 'Splunk',
               'messages__log_entries__service__self': 'P7RAELC',
               'messages__log_entries__channel__type': 'timeout',
               'messages__incident__last_status_change_by__id': 'ELC',
               'messages__incident__impacted_services__id': 'AELC',
               'messages__webhook__accounts_addon': None,
               'messages__incident__impacted_services__html_url': 'AELC',
               'messages__incident__teams__self': '59ZN5',
               'messages__incident__incident_key': None,
               'messages__webhook__outbound_integration__html_url': None,
               'messages__incident__service__alert_grouping_timeout': None,
               'messages__incident__escalation_policy__id': 'S2R6',
               'messages__incident__service__teams__self': 'P059ZN5',
               'messages__incident__service__incident_urgency_rule__type': 'constant',
               'messages__incident__service__support_hours': None,
               'messages__webhook__name': 'Data Pipeline Integration Queue',
               'messages__incident__first_trigger_log_entry__html_url': 'BHVZH38',
               'messages__log_entries__teams__summary': 'Card-TSYS-ApplePay',
               'messages__incident__acknowledgements': None,
               'messages__incident__teams__summary': 'Card',
               'messages__incident__service__integrations__id': 'SEA',
               'messages__log_entries__contexts': None,
               'messages__incident__service__alert_creation':
                  'create_alerts_and_incidents',
               'messages__id': 'f73f-11e8-af44-0a5ff7c88466',
               'messages__incident__service__response_play': None,
               'messages__incident__service__integrations__html_url': 'SEA',
               'messages__incident__service__metadata__conference_url': '',
               'messages__incident__first_trigger_log_entry__summary':
                  'Triggered through the API',
               'messages__incident__service__teams__html_url': 'P059ZN5',
               'messages__incident__service__id': 'ELC',
               'messages__incident__service__escalation_policy__summary': 'ApplePay',
               'messages__webhook__html_url': None,
               'messages__log_entries__teams__type': 'team_reference',
               'messages__incident__responder_requests': None,
               'messages__webhook__type': 'webhook',
               'messages__log_entries__teams__self': 'P059ZN5',
               'messages__incident__type': 'incident',
               'messages__incident__assignments__at': '2018-12-03T20:58:58Z',
               'messages__incident__service__escalation_policy__type':
                  'escalation_policy_reference',
               'messages__incident__last_status_change_by__summary': 'Card',
               'messages__log_entries__self': 'TTEV',
               'messages__incident__service__alert_grouping': None,
               'messages__log_entries__service__id': 'P7RAELC',
               'messages__incident__service__last_incident_timestamp':
                  '2018-12-03T16:08:55-05:00',
               'messages__incident__service__created_at': '2018-04-10T18:05:25-04:00',
               'messages__incident__service__metadata__conference_number': '',
               'messages__webhook__id': 'P9S8SCO',
               'messages__log_entries__agent__id': '7RAELC',
               'messages__incident__incident_number': 2317252,
               'messages__incident__assignments__assignee__self': '',
               'messages__incident__service__addons': None,
               'messages__log_entries__summary': 'Resolved by timeout',
               'messages__incident__service__status': 'critical',
               'messages__log_entries__service__type': 'service_reference',
               'messages__incident__service__teams__id': 'P059ZN5',
               'messages__incident__service__self': 'ELC',
               'messages__incident__subscriber_requests': None,
               'messages__incident__service__description': 'MANAGEMENT',
               'messages__log_entries__agent__summary': 'Card',
               'messages__webhook__webhook_object__id': 'P7RAELC',
               'messages__webhook__description': None,
               'messages__incident__created_at': '2018-12-03T20:58:58Z',
               'messages__incident__summary': 'TSYS',
               'messages__incident__first_trigger_log_entry__id': 'BHVZH38',
               'messages__log_entries__teams__id': 'P059ZN5',
               'messages__incident__service__type': 'service',
               'messages__incident__impacted_services__summary': 'Card',
               'messages__incident__last_status_change_by__html_url': 'ELC',
               'messages__log_entries__agent__type': 'service_reference',
               'messages__incident__alert_grouping': None,
               'messages__incident__impacted_services__type': 'service_reference',
               'messages__incident__resolve_reason': None,
               'messages__incident__self': 'BK744Y',
               'messages__incident__impacted_services__self': 'AELC',
               'messages__incident__html_url': '744Y',
               'messages__incident__urgency': 'low',
               'messages__log_entries__incident__summary': ' IDs',
               'messages__log_entries__type': 'resolve_log_entry',
               'messages__created_on': '2018-12-03T21:08:58Z',
               'messages__webhook__endpoint_url': '',
               'messages__incident__description': 'IDs',
               'messages__incident__service__integrations__self': 'SEA',
               'messages__log_entries__service__html_url': 'P7RAELC',
               'messages__log_entries__incident__self': 'PBK744Y',
               'messages__log_entries__incident__type': 'incident_reference',
               'messages__incident__service__html_url': 'ELC',
               'messages__webhook__outbound_integration__summary': 'Generic V2 Webhook',
               'messages__incident__priority': None,
               'messages__webhook__webhook_object__self': 'P7RAELC',
               'messages__webhook__webhook_object__summary': 'Card',
               'messages__incident__assignments__assignee__type': 'user_reference',
               'messages__log_entries__service__summary':
                  'Card-TSYS-ApplePay-ENVPRTSYSPLATFORMMANAGEMENT',
               'messages__log_entries__teams__html_url': 'P059ZN5',
               'messages__webhook__outbound_integration__id': 'PJFWPEP',
               'messages__log_entries__html_url': None,
               'messages__incident__last_status_change_by__self': 'ELC',
               'messages__incident__service__escalation_policy__self': 'R6',
               'messages__incident__id': '744Y',
               'messages__incident__escalation_policy__summary': 'ApplePay',
               'messages__incident__service__auto_resolve_timeout': 600,
               'messages__incident__is_mergeable': True,
               'messages__incident__service__incident_urgency_rule__urgency': 'low',
               'messages__webhook__summary': 'Data Pipeline Integration Queue',
               'messages__incident__last_status_change_at': '2018-12-03T21:08:58Z',
               'messages__incident__service__scheduled_actions': None,
               'messages__log_entries__incident__html_url': 'PBK744Y',
               'messages__webhook__self': 'P9S8SCO',
               'messages__log_entries__incident__id': 'PBK744Y',
               'messages__log_entries__created_at': '2018-12-03T21:08:58+00:00',
               'messages__incident__teams__id': '59ZN5',
               'messages__incident__escalation_policy__html_url': 'S2R6',
               'messages__log_entries__agent__html_url': '7RAELC',
               'messages__incident__alert_counts__resolved': 0,
               'messages__event': 'incident.resolve',
               'messages__incident__teams__type': 'team_reference',
               'messages__webhook__outbound_integration__self': 'PJFWPEP',
               'messages__incident__service__escalation_policy__html_url': 'R6',
               'messages__webhook__webhook_object__type': 'service_reference',
               'messages__incident__first_trigger_log_entry__type':
                  'trigger_log_entry_reference',
               'messages__incident__service__teams__summary': 'ApplePay',
               'messages__log_entries__agent__self': '7RAELC',
               'messages__webhook__outbound_integration__type':
                  'outbound_integration_reference',
               'messages__incident__assignments__assignee__summary': '',
               'messages__incident__external_references': None,
               'messages__incident__incidents_responders': None,
               'messages__incident__title': 'TSYS',
               'messages__incident__status': 'resolved'}])

    # Map types
    validate('rs_complex.map_t', [{'str_map__value': 'ab', 'id': 1, 'str_map__key': '1'}])
    validate('rs_complex.market_v20_single',
             [{'enterpriseeventenvelope__correlationid': None,
               'domainpayload__productid': '9998',
               'enterpriseeventenvelope__eventid': '83fe66b39ff3442b8a928f7299e43af6',
               'domainpayload__productdecisiontimestamp': '2018-04-03T15:12:17.194Z',
               'domainpayload__applicationid': '1522768336584',
               'domainpayload__externalkeys__key': 'cardCustomerRecognitionDecisionKey',
               'year': '2018',
               'domainpayload__decision': 'DECLINE',
               'domainpayload__turndownreasoncodes': None,
               'enterpriseeventenvelope__eventaction': 'Created',
               'enterpriseeventenvelope__eventprefix': '',
               'enterpriseeventenvelope__providerid': '40e840a6fce64ec78448baff73a14c04',
               'domainpayload__createdby': None,
               'enterpriseeventenvelope__eventdataquality__errorruleids': None,
               'enterpriseeventenvelope__domainpayloadversion': 'v1',
               'domainpayload__offercode': None,
               'enterpriseeventenvelope__eventqualifier': '',
               'domainpayload__concerntypes__value': None,
               'domainpayload__temporarycreditline': 0,
               'domainpayload__externalkeys__value': 'f4952756-0504-4b39-b178-b97f1dec2a6d',
               'domainpayload__creditline': 0,
               'domainpayload__productdecisionid': '83fe66b39ff3442b8a928f7299e43af6',
               'enterpriseeventenvelope__eventoccurrencetimestamp':
                  '2018-04-17T15:49:35.507',
               'domainpayload__concerntypes__key': None,
               'enterpriseeventenvelope__eventdataquality__executedrulesetid': '',
               'enterpriseeventenvelope__eventdataquality__failedruleids': None}])

    validate('rs_complex.market_v30_single',
             [{'domainpayload__temporarycreditline': 0,
               'enterpriseeventenvelope__eventdataquality__failedruleids':
                  '9566f8a1-97b2-42ef-94e8-2b8c38fb9d84',
               'domainpayload__createdby': None,
               'enterpriseeventenvelope__eventprefix': '',
               'enterpriseeventenvelope__providerid': '40e840a6fce64ec78448baff73a14c04',
               'domainpayload__decision': 'APPROVE',
               'year': '2018',
               'domainpayload__concerntypes__value': 'notRequired',
               'domainpayload__productdecisiontimestamp': '2018-11-02T02:00:31.973Z',
               'domainpayload__turndownreasoncodes': None,
               'domainpayload__externalkeys__value': '1f725222-30ab-4c1f-b39a-f1b60904046f',
               'domainpayload__creditline': 3000,
               'domainpayload__concerntypes__key': 'incomeVerification',
               'enterpriseeventenvelope__domainpayloadversion': 'v1',
               'enterpriseeventenvelope__eventid': '259bdeb898934724a5caf496a3a4ec9d',
               'enterpriseeventenvelope__eventdataquality__executedrulesetid':
                  '04ef8e1b-22ff-40de-8bdd-87bfc580e4b6',
               'domainpayload__offercode': '1280935',
               'domainpayload__productid': '3665',
               'domainpayload__applicationid': '10000057919105',
               'enterpriseeventenvelope__correlationid': None,
               'enterpriseeventenvelope__eventdataquality__errorruleids': None,
               'enterpriseeventenvelope__eventqualifier': '',
               'enterpriseeventenvelope__eventaction': 'Created',
               'enterpriseeventenvelope__eventoccurrencetimestamp':
                  '2018-11-02T02:00:31.977',
               'domainpayload__externalkeys__key': 'abilityToPayReferenceId',
               'domainpayload__productdecisionid': '259bdeb898934724a5caf496a3a4ec9d'}])
    validate('rs_complex.user_phone_numbers_map',
             [{'username': 'Alex', 'contact_numbers__value': '444-555-6666',
               'contact_numbers__key': 'work', 'uid': 1}])

    # Just flat or structs
    validate('rs_complex.bytes_type', [{'id': 1, 'v': 'abc'}])
    validate('rs_complex.bytes_type_file', [{'id': 1, 'v': 'abc'}])
    validate('rs_complex.enum_type', [{'id': 1, 'v': 'ENUM_1'}])
    validate('rs_complex.enum_type_default', [{'id': 1, 'v': 'ENUM_1'}])
    # This is non deterministic when selecting just 1 row
    #validate('rs_complex.market_decide_single_avro',
    #         [{'applicationid': '10000057919105', 'productid': '3665',
    #           'createdby': None, 'decision': 'APPROVE',
    #           'productdecisiontimestamp': '2018-11-02T02:00:31.973Z',
    #           'offercode': '1280935',
    #           'productdecisionid': '259bdeb898934724a5caf496a3a4ec9d', 'year': '2018'}])
    validate('rs_complex.struct_nested',
             [{'id': 123, 's1__f1': 'field1', 's1__s2__f3': '2', 's1__s2__f2': 2}])
    validate('rs_complex.struct_nested_s1',
             [{'s1__s2__f3': '2', 's1__f1': 'field1', 's1__s2__f2': 2}])
    validate('rs_complex.struct_nested_s1_f1', [{'f1': 'field1'}])
    validate('rs_complex.struct_nested_s1_s2', [{'s2__f2': 2, 's2__f3': '2'}])
    validate('rs_complex.struct_nested_view',
             [{'id': 123, 's1__s2__f2': 2, 's1__f1': 'field1', 's1__s2__f3': '2'}])
    validate('rs_complex.struct_t', [{'s1__f2': 1, 's1__f1': 'field1', 'id': 123}])
    validate('rs_complex.struct_t2',
             [{'id': 123, 's2__f1': 'field11', 's2__f2': 11,
               's1__f2': 1, 's1__f1': 'field1'}])
    validate('rs_complex.struct_t3',
             [{'s1__f1': 'field1', 's2__f2': 11,
               'id': 123, 's2__f1': 'field11', 's1__f2': 1}])
    validate('rs_complex.struct_t_id', [{'id': 123}])

    validate('rs_complex.struct_t_s1', [{'s1__f2': 1, 's1__f1': 'field1'}])
    validate('rs_complex.struct_t_view', [{'s1__f1': 'field1', 's1__f2': 1, 'id': 123}])
    validate('rs_complex.struct_t_view2', [{'id': 123, 's1__f2': 1, 's1__f1': 'field1'}])
    validate('rs_complex.users',
             [{'age': 25, 'uid': 100, 'address__city': 'san francisco',
               'user': 'alice', 'address__state': 'ca'}])

# Commented out ones are failing due to views, maps, or some other corner case.
# TODO: debug them
@pytest.mark.parametrize("unnest_dataset", [
    # View, not expected to work
    #'chase.cte_view',
    #'chase.redacted_ledger_balance_view',
    #'chase.subscription_party_view',

    # TODO: Something wrong with view.
    #'chase.subscription_view',
    #'rs_complex.strarray_t_view',
    #'rs_complex.view_over_multiple_structs',

    # TODO: Map
    #'rs_complex.map_t',
    #'rs_complex.market_v20_single',
    #'rs_complex.market_v30_single',
    #'rs_complex.user_phone_numbers_map',
    #'rs_complex_parquet.array_struct_map_t',
    #'rs_complex_parquet.map_array',
    #'rs_complex_parquet.map_array_t2',
    #'rs_complex_parquet.map_struct_array_t',
    #'rs_complex_parquet.map_struct_array_t_view',
    #'rs_complex_parquet.map_struct_t',
    #'rs_complex_parquet.map_struct_t_view',
    #'rs_complex_parquet.map_t',

    'chase.feescharges_view',
    'chase.ledger_balance',
    'chase.ledger_transaction',
    'chase.party',
    'chase.second_level',
    'chase.struct_t',
    'chase.struct_view',
    'chase.subscription',
    'chase.subscription_currency',

    # TODO: either hung or taking very long
    'chase.product',
    'chase.zd1211',
    'chase.zd1211_1',
    #'chase.zd1211_join_view',
    #'chase.zd1211_view',
    #'chase.zd1238',
    #'chase.zd1238_1',

    'chase.zd1238_2',
    'chase.zd1238_3',
    'chase.zd1238_4',
    'chase.zd1238_5',
    'chase.zd1238_6',
    'chase.zd1238_7',
    'chase.zd1238_8',
    'chase.zd1238_9',

    'rs_complex.array_struct_array',
    'rs_complex.array_struct_t',
    'rs_complex.array_t',
    'rs_complex.avrotbl',
    'rs_complex.bytes_type',
    'rs_complex.bytes_type_file',
    'rs_complex.enum_type',
    'rs_complex.enum_type_default',
    'rs_complex.market_decide_single_avro',
    'rs_complex.multiple_structs_nested',
    'rs_complex.strarray_t',
    'rs_complex.struct_array_struct',
    'rs_complex.struct_nested',
    'rs_complex.struct_nested_s1',
    'rs_complex.struct_nested_s1_f1',
    'rs_complex.struct_nested_s1_s2',
    'rs_complex.struct_nested_view',
    'rs_complex.struct_t',
    'rs_complex.struct_t2',
    'rs_complex.struct_t3',
    'rs_complex.struct_t_id',
    'rs_complex.struct_t_s1',
    'rs_complex.struct_t_view',
    'rs_complex.struct_t_view2',
    'rs_complex.user_phone_numbers',
    'rs_complex.users',
    'rs_complex_parquet.array_struct_array',
    'rs_complex_parquet.array_struct_t',
    'rs_complex_parquet.array_struct_t2',
    'rs_complex_parquet.array_t',
    'rs_complex_parquet.bytes_type',
    'rs_complex_parquet.enum_type',
    'rs_complex_parquet.strarray_t',
    'rs_complex_parquet.struct_array_struct',
    'rs_complex_parquet.struct_nested',
    'rs_complex_parquet.struct_nested_s1',
    'rs_complex_parquet.struct_nested_s1_f1',
    'rs_complex_parquet.struct_nested_s1_s2',
    'rs_complex_parquet.struct_nested_view',
    'rs_complex_parquet.struct_t',
    'rs_complex_parquet.struct_t2',
    'rs_complex_parquet.struct_t3',
    'rs_complex_parquet.struct_t_id',
    'rs_complex_parquet.struct_t_s1',
    'rs_complex_parquet.struct_t_view',
    'rs_complex_parquet.struct_t_view2',
    #'rs_complex_parquet.spark_all_mixed_compression',
    #'rs_complex_parquet.spark_gzip',
    #'rs_complex_parquet.spark_snappy',
    #'rs_complex_parquet.spark_snappy_part',
    #'rs_complex_parquet.spark_uncompressed',
    #'rs_complex_parquet.spark_uncompressed_legacy_format',
    'customer.zd277_complex',
    'rs_json_format.complex_c1_case_sensitive',
    'rs_json_format.complex_c1_usecase',
    'rs_json_format.complex_nike_usecase',
    'rs_json_format.json_array_struct',
    'rs_json_format.json_arrays_test',
    'rs_json_format.json_inferred',
    'rs_json_format.json_primitives',
    'rs_json_format.json_primitives_inferred',
    'rs_json_format.json_struct',
    'rs_json_format.json_struct_array',
    'rs_json_format.json_struct_nested',
    'rs_json_format.primitives_with_array',
])

def test_view_full_unnest(unnest_dataset):
    print("Testing: " + unnest_dataset)
    ctx = common.get_test_context()
    db = 'unnest_test_db'

    def contains(col, attrs):
        if col not in attrs:
            return False
        return '%s.a1' % db in attrs[col]

    with common.get_planner(ctx) as conn:
        conn.execute_ddl("drop attribute if exists %s.a1" % db)
        conn.execute_ddl("create attribute %s.a1" % db)

        parts = unnest_dataset.split('.')
        unnest_scan = get_scan_as_json(conn, 'select ** from %s' % unnest_dataset)

        # Run a few iterations
        for _ in range(0, 3):
            # Add an attribute to a random leaf col
            base_path, flattened = common.TestBase.get_random_leaf_column(
                conn, parts[0], parts[1])
            print("Assigning tag to %s --> %s " % (base_path, flattened))
            conn.assign_attribute(db, 'a1', parts[0], parts[1], base_path)
            attrs_by_col = common.TestBase.collect_column_attributes(
                conn.list_datasets(parts[0], name=parts[1])[0])
            print(conn.execute_ddl_table_output('describe %s' % unnest_dataset))
            assert contains(unnest_dataset + '.' + base_path, attrs_by_col)

            # Create a view and make sure the scans are the same
            conn.execute_ddl("CREATE DATABASE IF NOT EXISTS %s" % db)
            conn.execute_ddl("DROP VIEW IF EXISTS %s.v" % db)
            conn.execute_ddl("CREATE VIEW %s.v AS SELECT ** from %s" %
                             (db, unnest_dataset))
            view_scan = get_scan_as_json(conn, 'select * from %s.v' % db)
            assert format(unnest_scan) == format(view_scan)

            # Verify tag inheritance
            print(conn.execute_ddl_table_output('describe %s.v' % db))
            attrs_by_col = common.TestBase.collect_column_attributes(
                conn.list_datasets(db, name='v')[0])
            assert contains(db + '.v.' + flattened, attrs_by_col)

            # Unassign the tag from the base table, should also be removed from view
            conn.unassign_attribute(db, 'a1', parts[0], parts[1], base_path, cascade=True)
            attrs_by_col = common.TestBase.collect_column_attributes(
                conn.list_datasets(parts[0], name=parts[1])[0])
            print(attrs_by_col)
            assert not contains(unnest_dataset + '.' + base_path, attrs_by_col)

            attrs_by_col = common.TestBase.collect_column_attributes(
                conn.list_datasets(db, name='v')[0])
            print(attrs_by_col)
            assert not contains(db + '.v.' + flattened, attrs_by_col)

# pylint: disable=line-too-long
def test_complex_json_format():
    assert_scan_output(
        'select * from rs_json_format.gharch_test_data where id = \'1201654257\'',
        [{'actor':
            {'avatar_url':
            b'https://secure.gravatar.com/avatar/1641c0f988b844f44de596fcef3adc62?d=http://github.dev%2Fimages%2Fgravatars%2Fgravatar-user-420.png',
            'gravatar_id': b'1641c0f988b844f44de596fcef3adc62',
            'id': 75915,
            'url': b'https://api.github.dev/users/mitechie',
            'login': b'mitechie'},
            'id': b'1201654257',
            'public': True,
            'payload': {
                'actor': b'mitechie', 'size': 1,
                'head': b'4d1085635d642ba60071341dd46d38216bff1313', 'push_id': 27295992,
                'actor_gravatar': b'1641c0f988b844f44de596fcef3adc62',
                'ref': b'refs/heads/develop',
                'shas':
                    [[
                        b'4d1085635d642ba60071341dd46d38216bff1313',
                        b'86d2a548fee49abc43bd1c251300d7089748195b@mitechie.com',
                        b'Garden it up',
                        b'Richard Harding'
                    ]],
                'repo': b'mitechie/Bookie'
            },
            'type': b'PushEvent',
            'repo':
                {'id': 1176073,
                'url': b'https://api.github.dev/repos/mitechie/Bookie',
                'name': b'mitechie/Bookie'},
            'created_at': '2011-03-23 00:06:16.000'
        }]
    )
# pylint: enable=bad-continuation
# pylint: enable=line-too-long



#pylint: skip-file

import base64
import hashlib
import hmac
import random
import struct
import sys

from okera._pure_sasl import SASLError, SASLProtocolException, QOP

try:
    import kerberos
    have_kerberos = True
except ImportError:
    have_kerberos = False

PY3 = sys.version_info[0] == 3
if PY3:
    def _b(s):
        return s.encode("utf-8")
else:
    def _b(s):
        return s


class Mechanism(object):
    """
    The base class for all mechanisms.
    """

    name = None
    """ The IANA registered name for the mechanism. """

    score = 0
    """ A relative security score where higher scores correspond
    to more secure mechanisms. """

    complete = False
    """ Set to True when SASL negotiation has completed succesfully. """

    has_initial_response = False

    allows_anonymous = True
    """ True if the mechanism allows for anonymous logins. """

    uses_plaintext = True
    """ True if the mechanism transmits sensitive information in plaintext. """

    active_safe = False
    """ True if the mechanism is safe against active attacks. """

    dictionary_safe = False
    """ True if the mechanism is safe against passive dictionary attacks. """

    qops = [QOP.AUTH]
    """ QOPs supported by the Mechanism """

    qop = QOP.AUTH
    """ Selected QOP """

    def __init__(self, sasl, **props):
        self.sasl = sasl

    def process(self, challenge=None):
        """
        Process a challenge request and return the response.

        :param challenge: A challenge issued by the server that
                          must be answered for authentication.
        """
        raise NotImplementedError()

    def wrap(self, outgoing):
        """
        Wrap an outgoing message intended for the SASL server. Depending
        on the negotiated quality of protection, this may result in the
        message being signed, encrypted, or left unaltered.
        """
        raise NotImplementedError()

    def unwrap(self, incoming):
        """
        Unwrap a message from the SASL server. Depending on the negotiated
        quality of protection, this may check a signature, decrypt the message,
        or leave the message unaltered.
        """
        raise NotImplementedError()

    def dispose(self):
        """
        Clear all sensitive data, such as passwords.
        """
        pass

    def _fetch_properties(self, *properties):
        """
        Ensure this mechanism has the needed properties. If they haven't
        been set yet, the registered callback function will be called for
        each property to retrieve a value.
        """
        needed = [p for p in properties if getattr(self, p, None) is None]
        if needed and not self.sasl.callback:
            raise SASLError('The following properties are required, but a '
                            'callback has not been set: %s' % ', '.join(needed))

        for prop in needed:
            setattr(self, prop, self.sasl.callback(prop))

    def _pick_qop(self, server_qop_set):
        """
        Choose a quality of protection based on the user's requirements,
        what the server supports, and what the mechanism supports.
        """
        user_qops = set(_b(qop) if isinstance(qop, str) else qop for qop in self.sasl.qops)  # normalize user-defined config
        supported_qops = set(self.qops)
        available_qops = user_qops & supported_qops & server_qop_set
        if not available_qops:
            user = b', '.join(user_qops).decode('ascii')
            supported = b', '.join(supported_qops).decode('ascii')
            offered = b', '.join(server_qop_set).decode('ascii')
            raise SASLProtocolException("Your requested quality of "
                                        "protection is one of (%s), the server is "
                                        "offering (%s), and %s supports (%s)" % (user, offered, self.name, supported))
        else:
            for qop in (QOP.AUTH_CONF, QOP.AUTH_INT, QOP.AUTH):
                if qop in available_qops:
                    self.qop = qop
                    break


class AnonymousMechanism(Mechanism):
    """
    An anonymous user login mechanism.
    """
    name = 'ANONYMOUS'
    score = 0

    uses_plaintext = False

    def process(self, challenge=None):
        self.complete = True
        return b'Anonymous, None'


class PlainMechanism(Mechanism):
    """
    A plaintext user/password based mechanism.
    """
    name = 'PLAIN'
    score = 1

    allows_anonymous = False

    def wrap(self, outgoing):
        return outgoing

    def unwrap(self, incoming):
        return incoming

    def __init__(self, sasl, username=None, password=None, identity='', **props):
        Mechanism.__init__(self, sasl)
        self.identity = identity
        self.username = username
        self.password = password

    def process(self, challenge=None):
        self._fetch_properties('username', 'password')
        self.complete = True
        auth_id = self.sasl.authorization_id or self.identity
        return b''.join((_b(auth_id), b'\x00', _b(self.username), b'\x00', _b(self.password)))

    def dispose(self):
        self.password = None


class ExternalMechanism(Mechanism):
    """
    The EXTERNAL mechanism allows a client to request the server to use
    credentials established by means external to the mechanism to
    authenticate the client.
    """
    name = 'EXTERNAL'
    score = 10

    def wrap(self, outgoing):
        return outgoing

    def unwrap(self, incoming):
        return incoming

    def process(self, challenge=None):
        return b''


class CramMD5Mechanism(PlainMechanism):
    name = "CRAM-MD5"
    score = 20

    allows_anonymous = False
    uses_plaintext = False

    def __init__(self, sasl, username=None, password=None, **props):
        Mechanism.__init__(self, sasl)
        self.username = username
        self.password = password

    def process(self, challenge=None):
        if challenge is None:
            return None

        self._fetch_properties('username', 'password')
        mac = hmac.HMAC(key=_b(self.password), digestmod=hashlib.md5)
        mac.update(challenge)
        self.complete = True
        return b''.join((_b(self.username), b' ', _b(mac.hexdigest())))

    def dispose(self):
        self.password = None


# functions used in DigestMD5 which were originally defined in the now-removed
# util module

def to_bytes(text):
    """
    Convert Unicode text to UTF-8 encoded bytes.

    Since Python 2.6+ and Python 3+ have similar but incompatible
    signatures, this function unifies the two to keep code sane.

    :param text: Unicode text to convert to bytes
    :rtype: bytes (Python3), str (Python2.6+)
    """
    if sys.version_info < (3, 0):
        import __builtin__
        return __builtin__.bytes(text)
    else:
        import builtins
        if isinstance(text, builtins.bytes):
            # We already have bytes, so do nothing
            return text
        if isinstance(text, list):
            # Convert a list of integers to bytes
            return builtins.bytes(text)
        else:
            # Convert UTF-8 text to bytes
            return builtins.bytes(str(text), encoding='utf-8')


def quote(text):
    """
    Enclose in quotes and escape internal slashes and double quotes.

    :param text: A Unicode or byte string.
    """
    text = to_bytes(text)
    return b'"' + text.replace(b'\\', b'\\\\').replace(b'"', b'\\"') + b'"'


class DigestMD5Mechanism(Mechanism):

    name = "DIGEST-MD5"
    score = 30

    allows_anonymous = False
    uses_plaintext = False

    def __init__(self, sasl, username=None, password=None, **props):
        Mechanism.__init__(self, sasl)
        self.username = username
        self.password = password

        self._digest_uri = None
        self._a1 = None

    def dispose(self):
        self._digest_uri = None
        self._a1 = None

        self.password = None
        self.key_hash = None
        self.realm = None
        self.nonce = None
        self.cnonce = None
        self.nc = 0

    def wrap(self, outgoing):
        return outgoing

    def unwrap(self, incoming):
        return incoming

    def response(self):
        required_props = ['username']
        if not getattr(self, 'key_hash', None):
            required_props.append('password')
        self._fetch_properties(*required_props)

        resp = {}
        resp['qop'] = self.qop

        if getattr(self, 'realm', None) is not None:
            resp['realm'] = quote(self.realm)

        resp['username'] = quote(to_bytes(self.username))
        resp['nonce'] = quote(self.nonce)
        if self.nc == 0:
            self.cnonce = to_bytes('%s' % random.random())[2:]
        resp['cnonce'] = quote(self.cnonce)
        self.nc += 1
        resp['nc'] = to_bytes('%08x' % self.nc)

        self._digest_uri = (
                to_bytes(self.sasl.service) + b'/' + to_bytes(self.sasl.host))
        resp['digest-uri'] = quote(self._digest_uri)

        a2 = b'AUTHENTICATE:' + self._digest_uri
        if self.qop != QOP.AUTH:
            a2 += b':00000000000000000000000000000000'
            resp['maxbuf'] = b'16777215'  # 2**24-1
        resp['response'] = self.gen_hash(a2)
        return b','.join(
            [
                to_bytes(k) + b'=' + to_bytes(v)
                for k, v in resp.items()
            ]
        )

    @staticmethod
    def parse_challenge(challenge):
        """Parse a digest challenge message.

        :param ``bytes`` challenge:
            Challenge message from the server, in bytes.
        :returns:
            ``dict`` of ``str`` keyword to ``bytes`` values.
        """
        ret = {}
        var = b''
        val = b''
        in_var = True
        in_quotes = False
        new = False
        escaped = False
        for c in challenge:
            if sys.version_info[0] == 3:
                c = to_bytes([c])
            if in_var:
                if c.isspace():
                    continue
                if c == b'=':
                    in_var = False
                    new = True
                else:
                    var += c
            else:
                if new:
                    if c == b'"':
                        in_quotes = True
                    else:
                        val += c
                    new = False
                elif in_quotes:
                    if escaped:
                        escaped = False
                        val += c
                    else:
                        if c == b'\\':
                            escaped = True
                        elif c == b'"':
                            in_quotes = False
                        else:
                            val += c
                else:
                    if c == b',':
                        if var:
                            ret[var.decode('ascii')] = val
                        var = b''
                        val = b''
                        in_var = True
                    else:
                        val += c
        if var:
            ret[var.decode('ascii')] = val
        return ret

    def gen_hash(self, a2):
        if not getattr(self, 'key_hash', None):
            key_hash = hashlib.md5()
            user = to_bytes(self.username)
            password = to_bytes(self.password)
            realm = to_bytes(self.realm)
            kh = user + b':' + realm + b':' + password
            key_hash.update(kh)
            self.key_hash = key_hash.digest()

        a1 = hashlib.md5(self.key_hash)
        a1h = b':' + self.nonce + b':' + self.cnonce
        a1.update(a1h)
        response = hashlib.md5()
        self._a1 = a1.digest()
        rv = to_bytes(a1.hexdigest().lower())
        rv += b':' + self.nonce
        rv += b':' + to_bytes('%08x' % self.nc)
        rv += b':' + self.cnonce
        rv += b':' + self.qop
        rv += b':' + to_bytes(hashlib.md5(a2).hexdigest().lower())
        response.update(rv)
        return to_bytes(response.hexdigest().lower())

    def authenticate_server(self, cmp_hash):
        a2 = b':' + self._digest_uri
        if self.qop != QOP.AUTH:
            a2 += b':00000000000000000000000000000000'
        if self.gen_hash(a2) != cmp_hash:
            raise SASLProtocolException('Invalid server auth response')

    def process(self, challenge=None):
        if challenge is None:
            needed = ['username', 'realm', 'nonce', 'key_hash',
                      'nc', 'cnonce', 'qops']
            if all(getattr(self, p, None) is not None for p in needed):
                return self.response()
            else:
                return None

        challenge_dict = DigestMD5Mechanism.parse_challenge(challenge)
        if 'rspauth' in challenge_dict:
            self.authenticate_server(challenge_dict['rspauth'])
            self.complete = True
            return None

        if 'realm' not in challenge_dict:
            self._fetch_properties('realm')
            challenge_dict['realm'] = self.realm

        for key in ('nonce', 'realm'):
            # TODO: rfc2831#section-2.1.1 realm: "Multiple realm directives are
            # allowed, in which case the user or client must choose one as the
            # realm for which to supply to username and password"
            # TODO: rfc2831#section-2.1.1 nonce: "This directive is required
            # and MUST appear exactly once; if not present, or if multiple
            # instances are present, the client should abort the authentication
            # exchange"
            if key in challenge_dict:
                setattr(self, key, challenge_dict[key])

        self.nc = 0
        if 'qop' in challenge_dict:
            server_offered_qops = [
                x.strip() for x in challenge_dict['qop'].split(b',')
            ]
        else:
            server_offered_qops = [QOP.AUTH]
        self._pick_qop(set(server_offered_qops))

        if 'maxbuf' in challenge_dict:
            self.max_buffer = min(
                self.sasl.max_buffer, int(challenge_dict['maxbuf']))

        # TODO: rfc2831#section-2.1.1 algorithm: This directive is required and
        # MUST appear exactly once; if not present, or if multiple instances
        # are present, the client should abort the authentication exchange.
        return self.response()


class GSSAPIMechanism(Mechanism):
    name = 'GSSAPI'
    score = 100
    qops = QOP.all

    allows_anonymous = False
    uses_plaintext = False
    active_safe = True

    def __init__(self, sasl, principal=None, **props):
        Mechanism.__init__(self, sasl)
        self.user = None
        self._have_negotiated_details = False
        self.host = self.sasl.host
        self.service = self.sasl.service
        self.principal = principal
        self._fetch_properties('host', 'service')

        krb_service = '@'.join((self.service, self.host))
        try:
            _, self.context = kerberos.authGSSClientInit(service=krb_service,
                                                         principal=self.principal)
        except TypeError:
            if self.principal is not None:
                raise Exception("Error: kerberos library does not support principal.")
            _, self.context = kerberos.authGSSClientInit(service=krb_service)

    def process(self, challenge=None):
        if not self._have_negotiated_details:
            kerberos.authGSSClientStep(self.context, '')
            _negotiated_details = kerberos.authGSSClientResponse(self.context)
            self._have_negotiated_details = True
            return base64.b64decode(_negotiated_details)

        challenge = base64.b64encode(challenge).decode('ascii')  # kerberos methods expect strings, not bytes
        if self.user is None:
            ret = kerberos.authGSSClientStep(self.context, challenge)
            if ret == kerberos.AUTH_GSS_COMPLETE:
                self.user = kerberos.authGSSClientUserName(self.context)
                return b''
            else:
                response = kerberos.authGSSClientResponse(self.context)
                if response:
                    response = base64.b64decode(response)
                else:
                    response = b''
            return response

        kerberos.authGSSClientUnwrap(self.context, challenge)
        data = kerberos.authGSSClientResponse(self.context)
        plaintext_data = base64.b64decode(data)
        if len(plaintext_data) != 4:
            raise SASLProtocolException("Bad response from server")  # todo: better message

        word, = struct.unpack('!I', plaintext_data)
        qop_bits = word >> 24
        max_length = word & 0xffffff
        server_offered_qops = QOP.names_from_bitmask(qop_bits)
        self._pick_qop(server_offered_qops)

        self.max_buffer = min(self.sasl.max_buffer, max_length)

        """
        byte 0: the selected qop. 1==auth, 2==auth-int, 4==auth-conf
        byte 1-3: the max length for any buffer sent back and forth on
            this connection. (big endian)
        the rest of the buffer: the authorization user name in UTF-8 -
            not null terminated.
        """
        auth_id = self.sasl.authorization_id or self.user
        l = len(auth_id)
        fmt = '!I' + str(l) + 's'
        word = QOP.flag_from_name(self.qop) << 24 | self.max_buffer
        out = struct.pack(fmt, word, _b(auth_id),)

        encoded = base64.b64encode(out).decode('ascii')

        kerberos.authGSSClientWrap(self.context, encoded)
        response = kerberos.authGSSClientResponse(self.context)
        self.complete = True
        return base64.b64decode(response)

    def wrap(self, outgoing):
        if self.qop != QOP.AUTH:
            outgoing = base64.b64encode(outgoing)
            if self.qop == QOP.AUTH_CONF:
                protect = 1
            else:
                protect = 0
            kerberos.authGSSClientWrap(self.context, outgoing, None, protect)
            return base64.b64decode(kerberos.authGSSClientResponse(self.context))
        else:
            return outgoing

    def unwrap(self, incoming):
        if self.qop != QOP.AUTH:
            incoming = base64.b64encode(incoming).decode('ascii')
            kerberos.authGSSClientUnwrap(self.context, incoming)
            conf = kerberos.authGSSClientResponseConf(self.context)
            if 0 == conf and self.qop == QOP.AUTH_CONF:
                raise Exception("Error: confidentiality requested, but not honored by the server.")
            return base64.b64decode(kerberos.authGSSClientResponse(self.context))
        else:
            return incoming

    def dispose(self):
        kerberos.authGSSClientClean(self.context)


#: Global registry mapping mechanism names to implementation classes.
mechanisms = dict((m.name, m) for m in (
    AnonymousMechanism,
    PlainMechanism,
    ExternalMechanism,
    CramMD5Mechanism,
    DigestMD5Mechanism))

if have_kerberos:
    mechanisms[GSSAPIMechanism.name] = GSSAPIMechanism



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for auth in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=protected-access

import unittest
import pytest
import prestodb

import thriftpy
from okera import context
from okera.tests import pycerebro_test_common as common

#
# NOTE: this test suite uses the fact that enable_token_auth will treat
# a string that has a period in it as a JWT/OkeraToken, and any other string
# as if it was the username (for an unauthed server). Since the test runs
# against an unauthed server, a JWT/OkeraToken will always fail, causing
# us to go through the retry logic.
#

# {
#   "sub": "user1",
#   "claim2": "user2",
#   "claim3": "user3"
# }
TOKEN = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJ1c2VyMSI' + \
        'sImNsYWltMiI6InVzZXIyIiwiY2xhaW0zIjoidXNlcjMifQ.4QIs75Pm' + \
        'B_aRtTVebR97CH9EugdNGj9-UAppkxDq73U'

def jwt_token_fn():
    return TOKEN

# Global variable for attempts
attempts = 0

def bad_then_good_func():
    global attempts
    token = None
    if attempts == 0:
        token = "foo.bar"
    else:
        token = "foo"
    attempts += 1
    return token

def bad_twice_then_good_func():
    global attempts
    token = None
    if attempts < 2:
        token = "foo.bar"
    else:
        # should never be called
        raise Exception()
    attempts += 1
    return token

def good_then_bad_func():
    global attempts
    token = None
    if attempts == 0:
        token = "foo"
    else:
        # this should never be called
        raise Exception()
    attempts += 1
    return token

def always_bad():
    return "foo.bar"

class AuthTest(unittest.TestCase):
    @staticmethod
    def _reset_attempts():
        global attempts
        attempts = 0

    def setUp(self):
        import requests
        import urllib3
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        AuthTest._reset_attempts()

    def test_token_func_basic(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_func=bad_then_good_func)
        with common.get_planner(ctx) as conn:
            results = conn.scan_as_json('okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'foo')

    def test_token_func_no_failure(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_func=bad_then_good_func)
        with common.get_planner(ctx) as conn:
            results = conn.scan_as_json('okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'foo')

    def test_token_func_and_token_str(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_func=always_bad, token_str='foo')
        with common.get_planner(ctx) as conn:
            results = conn.scan_as_json('okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'foo')

    def test_token_func_never_succeed(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_func=always_bad)
        with pytest.raises(thriftpy.transport.TTransportException):
            with common.get_planner(ctx):
                # we should never get here
                raise Exception()

    def test_token_func_retry_once_still_fail(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_func=bad_twice_then_good_func)
        with pytest.raises(thriftpy.transport.TTransportException):
            with common.get_planner(ctx):
                # we should never get here
                raise Exception()

    def test_token_func_non_picklable(self):
        def fn():
            return "foo"

        ctx = context()
        with pytest.raises(ValueError):
            ctx.enable_token_auth(token_func=fn)

    def test_token_user_extract(self):

        ctx = context()
        ctx.enable_token_auth(TOKEN)
        assert ctx._get_user() == 'user1'

        ctx.enable_token_auth(TOKEN, user_claims=['claim2'])
        assert ctx._get_user() == 'user2'

        ctx.enable_token_auth(TOKEN, user_claims=['nonexistent', 'claim2'])
        assert ctx._get_user() == 'user2'

        ctx.enable_token_auth(TOKEN, user_claims=['claim3', 'claim2'])
        assert ctx._get_user() == 'user3'

        ctx.enable_token_auth(TOKEN, user_claims=['nonexistent1', 'nonexistent2'])
        assert ctx._get_user() == 'user1'

        ctx.enable_token_auth(TOKEN, user_claims=[])
        assert ctx._get_user() == 'user1'

        # It's hard to test the internal logic for re-fetching
        # tokens when they are JWTs since currently Okera tests
        # run in un-authed mode. Due to that, we use the internal
        # functions to verify the logic is correct.

        ctx.enable_token_auth(token_func=jwt_token_fn, user_claims=['claim2'])
        assert ctx._get_user() == 'user2'
        ctx._generate_token()
        assert ctx._get_user() == 'user2'

    def test_presto_token_func_basic(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx, dialect='presto') as conn:
            ctx.enable_token_auth(token_func=bad_then_good_func, user='foo')
            results = conn.scan_as_json('select * from okera_sample.whoami', dialect='presto')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'foo')

    def test_presto_token_func_and_token_str(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx, dialect='presto') as conn:
            ctx.enable_token_auth(token_func=always_bad, token_str='foo')
            results = conn.scan_as_json('select * from okera_sample.whoami', dialect='presto')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'foo')

    def test_presto_token_func_never_succeed(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx, dialect='presto') as conn:
            ctx.enable_token_auth(token_func=always_bad, user='foo')
            with pytest.raises(prestodb.exceptions.DatabaseError):
                results = conn.scan_as_json('select * from okera_sample.whoami', dialect='presto')
                # we should never get here
                raise Exception()

    def test_presto_token_func_retry_once_still_fail(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx, dialect='presto') as conn:
            ctx.enable_token_auth(token_func=bad_twice_then_good_func, user='foo')
            with pytest.raises(prestodb.exceptions.DatabaseError):
                results = conn.scan_as_json('select * from okera_sample.whoami', dialect='presto')
                # we should never get here
                raise Exception()
